#include <algorithm>
#include <iostream>
#include <ranges>
#include <cctype>

#include "AtmAdmin.h"
#include "Errorseal.h"
#include "XMLStorage.h"

namespace AtmAdmin {
    // Used to make certain functions more concise and easier to read.
    enum Choice {
        PIN,
        NAME,
        MONEY,
    };

    void handleInvalidArguments(int argc) {
        if (argc != 2) {
            std::cerr << "Invalid number of arguments: " << argc - 1 << '\n';

            if (argc < 2) {
                std::cerr << "The ability to make new accounts is not implemented." << '\n';
                std::cerr << "Please put in a pre-existing name." << std::endl;
            }
            else
                std::cerr << "Too many arguments." << std::endl;

            exit (EXIT_FAILURE);
        }
    }

    // Allows the user to double-check their inputted information.
    // It also saves and applies the information if the user accepts it.
    void confirmUserInformation (XMLStorage& mainStorage, const std::string& tempInformation, const Choice choice) {
        char userCheck;
        bool loopDone = true;

        do {
            if (choice == PIN)
                std::cout << "Your new pin will be: " << tempInformation << '\n';
            else if (choice == NAME)
                std::cout << "Your new name will be: " << tempInformation << '\n';
            std::cout << "Is this okay? (y or n) >: ";
            std::cin.get(userCheck);
            userCheck = static_cast<char>(std::tolower(userCheck));
            if (userCheck == 'y' and choice == PIN) {
                mainStorage.pin = stoi(tempInformation);
                std::cout << "New pin applied." << '\n';
                loopDone = true;
            }

            else if (userCheck == 'y' and choice == NAME) {
                mainStorage.name = tempInformation;
                std::cout << "New name applied." << '\n';
                loopDone = true;
            }

            else if (userCheck == 'n') {
                std::cout << "Discarding..." << '\n';
                loopDone = true;
            }

            else {
                std::cout << "Invalid choice." << "\n\n";
                loopDone = false;
            }

        } while (!loopDone);
    }

    // Controls the flow of the program based on the user's input.
    bool handleUserInformation(char choice, XMLStorage& mainStorage) {
        std::string tempInformation{};

        switch (choice) {
            case '1':
                // changeAccountPin is a function found later in the code.
                // It is both the input, and algorithm for the new pin.

                // The algorithm returns "invalid" when something isn't as it should be.
                // This is handled by the if statement.
                tempInformation = changeAccountPin();
                if (tempInformation != "invalid") {
                    confirmUserInformation(mainStorage, tempInformation, PIN);
                    return true;
                }
                break;
            case '2':
                // changeAccountName is a function found later in the code.
                // It is both the input, and algorithm for the new name.

                // The algorithm returns "invalid" when something isn't as it should be.
                // This is handled by the if statement.
                tempInformation = changeAccountName();

                if (tempInformation != "invalid") {
                    confirmUserInformation(mainStorage, tempInformation, NAME);
                    return true;
                }
                break;
            case '3':
                // Simply a way to cheat, very straight-forward.

                // The code below is a placeholder
                std::cout << "Enter the amount you wish to have: ";
                std::cin >> mainStorage.money;
                break;
            case 'q':
                return false;
            default:
                std::cout << "Invalid choice.\n";
        }
        return true;
    }

    // Allows for the quick and clean creation of error reports.
    void errorReport (std::string_view error) {
        std::cout << error << '\n';
        std::cout << "Press enter to continue...";
        std::cin.get();
    };

    std::string changeAccountPin() {
        std::string newPin{};

        std::cout << "Enter new pin: ";
        std::cin >> newPin;
        Error::ignoreLine();

        std::array<int, 4> pinDigits{};

        // Asserts that the pin is exactly 4 letters in length.
        if (newPin.size() != 4) {
            errorReport("The pin has to be 4 numbers long.");
            return "invalid";
        }

        // Fills the int array with a number version of the string array.
        for (size_t i{}; i != 4; i++) {
            pinDigits[i] = (newPin[i] - '0');
        }

        // Checks if anything put into the pin is alphabetic.
        for (size_t i = 0; i < 4; i++) {
            if (!isdigit(newPin[i])) {
                errorReport("The pin cannot contain alphabet letters.");
                return "invalid";
            }
        }

        // Checks for sequential numbers (examples: 1234, 6789)
        { // START OF BRACKET SCOPE //
            int sequentialCount {0};

            for (size_t i{0}; i < 3; i++) {
                if (pinDigits[i]+1 == pinDigits[i+1] or pinDigits[i]-1 == pinDigits[i+1])
                    sequentialCount += 1;


                if (sequentialCount == 3) {
                    errorReport("Pin is too simple.");
                    return "invalid";
                }
            }
        } // END OF BRACKET SCOPE //

        // Checks for too many duplicate numbers.
        { // START OF BRACKET SCOPE //
            std::string tempPin = newPin;
            std::ranges::sort(tempPin.begin(), tempPin.end());


            if ( (tempPin[0] == tempPin[1] and tempPin[1] == tempPin[2]) // Checks if the first 3 numbers are the same.
                 or
                 (tempPin[1] == tempPin[2] and tempPin[2] == tempPin[3]) // Checks if the last 3 numbers are the same.
                 or
                 (tempPin[0] == tempPin[1] and tempPin[2] == tempPin[3])) // Checks if the first 2 and last 2 numbers are the same.
            {
                errorReport("Too many duplicate numbers.");
                return "invalid";
            }
        } // END OF BRACKET SCOPE //

       return newPin;
    }

    std::string changeAccountName() {
        std::string newName{};


        std::cout << "Enter new name: ";
        Error::ignoreLine();
        std::getline(std::cin, newName);

        // Ensures the name is between 2 and 15 characters long.
        if (newName.length() < 2 or newName.length() > 15) {
            errorReport("Name has to be between 2 and 15 characters.");
            return "invalid";
        }

        // Checks if any spaces are in the name, if so, reject it.
        if ( std::ranges::any_of(newName.cbegin(), newName.cend(), [] (char input) {
            return input == ' ';
        })) {
            errorReport("Name cannot contain spaces.");
            return "invalid";
        }

        // Turns all the letters into lowercase.
        std::ranges::transform (newName.cbegin(), newName.cend(), newName.begin(), [] (unsigned char input) { return tolower(input); });
        return newName;
    }
}
